/* split.c -- split a file into pieces.
   Copyright (C) 88, 91, 95, 1996 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/* By tege@sics.se, with rms.

   To do:
   * Implement -t CHAR or -t REGEX to specify break characters other
     than newline. */

#include <config.h>

#include <stdio.h>
#include <unistd.h>   /* F"ur Mapping ... */
#include <getopt.h>
#include <sys/types.h>
#include <sys/mman.h>   /* F"ur Mapping ... */

#if HAVE_LIMITS_H
# include <limits.h>
#endif

#ifndef UINT_MAX
# define UINT_MAX ((unsigned int) ~(unsigned int) 0)
#endif

#ifndef INT_MAX
# define INT_MAX ((int) (UINT_MAX >> 1))
#endif

#include "system.h"
#include "error.h"
#include "xstrtol.h"

/* Warum das immer noch keine Standart-Definitionen sind weiss ich nicht .. */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

/* Laut man-Page sollte dieser Wert eigentlich definiert sein; 
 * wenn er es nicht ist, dann wird er hier gesetzt .... 
 * Ob das Tats"achlich so auf allen Systemen geht, ist eher zu bezweifeln
 * MREMAP_MAYMOVE Ist auch nicht "uberall vorhanden ....
*/

#ifndef MAP_FAILED
#define MAP_FAILED -1
#endif

#ifndef MREMAP_MAYMOVE
#define MREMAP_MAYMOVE 0  /* Da es bitweise geodert wird st"ort es so nicht */
#endif

char *xmalloc ();
int full_write ();
int safe_read ();

/* The name this program was run with. */
char *program_name;

/* Base name of output files.  */
static char *outfile;

/* Pointer to the end of the prefix in OUTFILE.
   Suffixes are inserted here.  */
static char *outfile_mid;

/* Pointer to the end of OUTFILE. */
static char *outfile_end;

/* Name of input file.  May be "-".  */
static char *infile;

/* Descriptor on which input file is open.  */
static int input_desc;

/* Descriptor on which output file is open.  */
static int output_desc;
static char *obp;
static size_t ob_size;

/* If nonzero, display usage information and exit.  */
static int show_help;

/* If nonzero, print the version on standard output then exit.  */
static int show_version;

/* If nonzero, print a diagnostic on standard error just before each
   output file is opened. */
static int verbose;

/* Mapping-Flags : Eines f"ur das input-file und eines f"ur das output file;
 * Es kann nicht davon ausgegangen werden, dass das output-file immer map-
 * bar ist; es k"onnte der Fall eintreten, dass nicht genug Speicher vor-
 * handen ist; dann darf der Aufruf nicht fehlschlagen, sondern muss seine
 * Aufgabe auf traditionelle Weise erledigen. */

static int in_map, out_map;


static struct option const longopts[] =
{
  {"bytes", required_argument, NULL, 'b'},
  {"lines", required_argument, NULL, 'l'},
  {"line-bytes", required_argument, NULL, 'C'},
  {"verbose", no_argument, NULL, 2},
  {"help", no_argument, &show_help, 1},
  {"version", no_argument, &show_version, 1},
  {NULL, 0, NULL, 0}
};

static void
usage (int status, const char *reason)
{
  if (reason != NULL)
    fprintf (stderr, "%s: %s\n", program_name, reason);

  if (status != 0)
    fprintf (stderr, _("Try `%s --help' for more information.\n"),
	     program_name);
  else
    {
      printf (_("\
Usage: %s [OPTION] [INPUT [PREFIX]]\n\
"),
	      program_name);
    printf (_("\
Output fixed-size pieces of INPUT to PREFIXaa, PREFIXab, ...; default\n\
PREFIX is `x'.  With no INPUT, or when INPUT is -, read standard input.\n\
\n\
  -C, --line-bytes=SIZE   put at most SIZE bytes of lines per output file\n\
  -b, --bytes=SIZE        put SIZE bytes per output file\n\
  -l, --lines=NUMBER      put NUMBER lines per output file\n\
      --verbose           print a diagnostic to standard error just\n\
			    before each output file is opened\n\
  -NUMBER                 same as -l NUMBER\n\
      --help              display this help and exit\n\
      --version           output version information and exit\n\
\n\
SIZE may have a multiplier suffix: b for 512, k for 1K, m for 1 Meg.\n\
"));
    }
  exit (status == 0 ? EXIT_SUCCESS : EXIT_FAILURE);
}

/* Compute the next sequential output file name suffix and store it
   into the string `outfile' at the position pointed to by `outfile_mid'.  */

static void
next_file_name (void)
{
  int x;
  char *ne;
  unsigned int i;

  static int first_call = 1;

  /* Status for outfile name generation.  */
  static unsigned outfile_count = 0;
  static unsigned outfile_name_limit = 25 * 26;
  static unsigned outfile_name_generation = 1;

  if (!first_call)
    outfile_count++;
  first_call = 0;
  if (outfile_count < outfile_name_limit)
    {
      for (ne = outfile_end - 1; ; ne--)
	{
	  x = *ne;
	  if (x != 'z')
	    break;
	  *ne = 'a';
	}
      *ne = x + 1;
      return;
    }

  outfile_count = 0;
  outfile_name_limit *= 26;
  outfile_name_generation++;
  *outfile_mid++ = 'z';
  for (i = 0; i <= outfile_name_generation; i++)
    outfile_mid[i] = 'a';
  outfile_end += 2;
}

/* Write BYTES bytes at BP to an output file.
   If NEW_FILE_FLAG is nonzero, open the next output file.
   Otherwise add to the same output file already in use.  */
/* Es wird auf alle F"alle versucht die Ausgabedateien in den Speicher 
 * einzublenden. Sollte das aus irgendeinem Grund nicht m"oglich sein, 
 * wird wieder auf den traditionellen write()-Aufruf zur"uckgegriffen.
 * Zu beachten ist, dass nach dem "Offnen(Erzeugen) der Datei ein 
 * ftruncate() mit der entsprechenden Gr"osse erfolgen muss, da ansonst
 * die gemapte Datei nicht gr"osser wird ! 
 */

static void
cwrite (int new_file_flag, const char *bp, int bytes)
{
  if (new_file_flag)                      /* Neues Ausgabefile erzeugen ... */
    {               /* Dazu muss das vorherige zuerst geschlossen werden .. */
      if (output_desc >= 0 && close (output_desc) < 0)
	error (EXIT_FAILURE, errno, "%s", outfile);      /* Ende bei Fehler */
      if (out_map)      /* Wenn die Datei gemapped war, erfogt ein unmap .. */
	{
	  if ((obp != NULL) && (munmap (obp, ob_size) != 0))
	    error (EXIT_FAILURE, errno, "%s", outfile);  /* Ende bei Fehler */
          else 
	    {
	      obp = NULL;                 /* R"ucksetzen der globalen Werte */
	      ob_size = 0;
	    }
	}
      next_file_name ();                    /* Neuen Dateinamen erzeugen .. */
      if (verbose)
	fprintf (stderr, _("creating file `%s'\n"), outfile);

      /* "Anderung von O_WRONLY auf O_RDWR ist erforderlich, da es nicht
       * m"oglich ist eine Datei die nur WRONLY ge"offnet wurde zu mappen.
       * Diese (sonderbare) Eigenschaft wurde schon in einer Tabelle fest-
       * gehalten .... */ 

      output_desc = open (outfile, O_RDWR | O_CREAT | O_TRUNC, 0666);
      if (output_desc < 0) 
	error (EXIT_FAILURE, errno, "%s", outfile);       /* Ende bei Fehler */
      /* Es wird auf jeden Fall versucht die Datei zu mappen ... */
      if (((obp = mmap(NULL, (size_t)bytes, PROT_WRITE, 
		      MAP_SHARED, output_desc, 0)) == (char *) MAP_FAILED) ||
	    ftruncate (output_desc, (size_t) (ob_size + bytes)) != 0)
	{ /* Es wird nur ein Hinweis ausgegeben und mit write() fortgesetzt. */
	  out_map = FALSE;
          fprintf (stderr, "Can't mmap output-file ...\n");
	}
      else                                      /* Datei ist eingeblendet .. */
	{
	  out_map = TRUE;
          ob_size = bytes;
	}
    }
  if ((out_map) && (! new_file_flag))    /* Weitere Ausgabe in gemapte Datei */
    {    /* erfordert eine Vergr"osserung der Ausgabedatei mit ftruncate ... */
      if (ftruncate (output_desc, (size_t) (ob_size + bytes)) != 0)
	{          /* Im Fehlerfall wird wieder traditionell vortgesetzt ... */
	  out_map = FALSE;
	  fprintf (stderr, "Can't ftruncate output-file ...\n");
	}
      else /* Die Vergr"osserung der Ausgabe-Datei erfordert ein neuerliches */
	{ /* Einblenden mit der neuen Gr"osse : mremap() .. 
	     Da diese Funktion nicht "uberall zur Verf"ugung steht, wird sie
	     durch eine munmap()-mmap()-Kombination ersetzt.                 */
	  munmap (obp, ob_size);
	  if ((obp = mmap(NULL, (size_t)(ob_size + bytes), PROT_WRITE, 
			  MAP_SHARED, output_desc, 0)) == (char *) MAP_FAILED)
	    {
	      out_map = FALSE;
	      fprintf (stderr, "Can't mmap (remap) output-file ...\n");
	    }
	  /*	  if ((obp = mremap(obp, ob_size , (size_t) (ob_size + bytes),
			    MAP_SHARED | MREMAP_MAYMOVE)) == NULL)
	    {
	      out_map = FALSE;
	      fprintf (stderr, "Can't mremap output-file ...\n"); 
	    } */
	  else ob_size += bytes;
	}
    }

  if (out_map)     /* Im Map-Fall einfaches Kopieren in die Ausgabe-Datei .. */
    {
    memcpy ((void *)&obp[ob_size - bytes], (const void *)bp, (size_t) bytes);
    }
  else if (full_write (output_desc, bp, bytes) < 0)
    error (EXIT_FAILURE, errno, "%s", outfile); /* Ende bei Fehler */
}

/* Read NCHARS bytes from the input file into BUF.
   Return the number of bytes successfully read.
   If this is less than NCHARS, do not call `stdread' again.  */

static int
stdread (char *buf, int nchars)
{
  int n_read;
  int to_be_read = nchars;

  while (to_be_read)
    {
      n_read = safe_read (input_desc, buf, to_be_read);
      if (n_read < 0)
	return -1;
      if (n_read == 0)
	break;
      to_be_read -= n_read;
      buf += n_read;
    }
  return nchars - to_be_read;
}

/* Split into pieces of exactly NCHARS bytes.
   Use buffer BUF, whose size is BUFSIZE.  */
/* F"ur dei Verwendung von Mapping wird ein eigener Zweig in die 
 * split-Funktionen eingebaut. */

static void
bytes_split (int nchars, char *buf, int bufsize)
{
  int n_read;
  int new_file_flag = 1;
  int to_read;
  int to_write = nchars;
  char *bp_out;


  /* Im Mapping-Fall ist buf der Zeiger auf die gemappte Datei und bufsize
   * die Gr"osse der Datei. Der Buffer ist in diesem Fall eben immer so
   * gross wie die Datei .... */
 
  if (in_map)
    {
      n_read = 0;        /* Gesamtanzahl der verarbeiteten Bytes .. */
      for (;;)
	{
	  if ((nchars + n_read) >= bufsize) /* Nur mehr ein Rest zu lesen .. */
	    {    /* letzter Aufruf */
	      cwrite (1, &buf[n_read], (bufsize - n_read)); 
	      return; 
	    }
	  cwrite (1, &buf[n_read], nchars); /* richtige Lese-Position !! */
	  n_read += nchars;
	}
    } /* Hier kommt man nicht mehr raus; Ausstieg "uber das return .. */

  do
    {
      n_read = stdread (buf, bufsize);
      if (n_read < 0)
        error (EXIT_FAILURE, errno, "%s", infile);
      bp_out = buf;
      to_read = n_read;
      for (;;)
	{
	  if (to_read < to_write)
	    {
	      if (to_read)	/* do not write 0 bytes! */
		{
		  cwrite (new_file_flag, bp_out, to_read);
		  to_write -= to_read;
		  new_file_flag = 0;
		}
	      break;
	    }
	  else
	    {
	      cwrite (new_file_flag, bp_out, to_write);
	      bp_out += to_write;
	      to_read -= to_write;
	      new_file_flag = 1;
	      to_write = nchars;
	    }
	}
    }
  while (n_read == bufsize);
}

/* Split into pieces of exactly NLINES lines.
   Use buffer BUF, whose size is BUFSIZE.  */

static void
lines_split (int nlines, char *buf, int bufsize)
{
  int n_read = 0;
  char *bp, *bp_out, *eob;
  int new_file_flag = 1;
  int n = 0;

  /* Im Mapping-Fall ist buf der Zeiger auf die gemappte Datei und bufsize
   * die Gr"osse der Datei. Der Buffer ist in diesem Fall eben immer so
   * gross wie die Datei .... */
 
  if (in_map)
    {
      register int i = 0;      /* Byte-Z"ahler - als Index um 1 weniger !! */
      bp = buf;
      for (;;)                         /* "Uber die ganze Datei/Puffer ... */
	{          /* Zuerst bis zum Zeilenende (oder Dateiende) lesen ... */
          i++;                 /* Index beim Zeilen"uberlesen erh"ohen ... */
          while (((i + n_read) < bufsize) && (bp[i-1] != '\n')) i++;
	  if (bp[i-1] == '\n') n++;              /* Zeilenz"ahler erh"ohen */
	  if (((i + n_read) >= bufsize) || (n == nlines))
	    {     /* Ausgabe wenn Zeilenanzahl oder Dateiende erreicht ist */
	      cwrite (1, bp, i);
	      if ((n_read += i) >= bufsize) return; /* Alles gelesen - Ende */
              bp = buf + n_read;
	      i = 0;                                    /* R"ucksetzen ... */
              n = 0;
	    }
	}
    }       /* Hier kommt man nicht mehr raus; Ausstieg "uber das return .. */
  do
    {
      n_read = stdread (buf, bufsize);
      if (n_read < 0)
	error (EXIT_FAILURE, errno, "%s", infile);
      bp = bp_out = buf;
      eob = bp + n_read;
      *eob = '\n';
      for (;;)
	{
	  while (*bp++ != '\n')
	    ;			/* this semicolon takes most of the time */
	  if (bp > eob)
	    {
	      if (eob != bp_out) /* do not write 0 bytes! */
		{
		  cwrite (new_file_flag, bp_out, eob - bp_out);
		  new_file_flag = 0;
		}
	      break;
	    }
	  else
	    if (++n >= nlines)
	      {
		cwrite (new_file_flag, bp_out, bp - bp_out);
		bp_out = bp;
		new_file_flag = 1;
		n = 0;
	      }
	}
    }
  while (n_read == bufsize);
}

/* Split into pieces that are as large as possible while still not more
   than NCHARS bytes, and are split on line boundaries except
   where lines longer than NCHARS bytes occur. */
/* F"ur den Mapping-Fall muss der Aufruf an das Format der beiden anderen
 * angepasst werden, da wieder die gemappte Datei als Puffer/Gr"osse-Paar 
 * gebraucht wird ... */

static void
/* line_bytes_split (int nchars)  */
line_bytes_split (int nchars, char *inp, int size)
{
  int n_read;
  char *bp;
  int eof = 0;
  int n_buffered = 0;
  char *buf;

  /* Im Mapping-Fall ist inp der Zeiger auf die gemappte Datei und size
   * die Gr"osse der Datei. Der Buffer ist in diesem Fall eben immer so
   * gross wie die Datei .... */

  if (in_map)
    {
      /* Bedeutungen ("Anderungen) der verwendeten Variablen :
       * inp    : Mapping (File-)Anfang
       * size   : File-L"ange
       * buf    : aktuelle Anfangsposition
       * bp     : aktuelle Endposition
       * n_read : bisher gelesen
       */
 
      n_read = 0;
      while (n_read < size) /* Solange das Dateiende nicht erreicht ist .. */
	{
        buf = inp + n_read;                           /* Aktueller Anfang  */
        bp = buf + nchars;                            /* Aktuelles Ende    */
        if ((bp - inp) >= size)              /* Das Fileende erkennen .... */
	  {
	    n_read = size;           /* Werte an das File-Ende anpassen .. */
            bp = inp + size;
	  }
        else                            /* mitten drinnen in der Datei ... */
	  {  /* von der Zeilengr"osse bis zum letzten Zeilenende zur"ck .. */
	    while (bp > buf && bp[-1] != '\n') bp--; 
            if (bp == buf)  bp = buf + nchars; /* keine Newline drinnen .. */
	    n_read += (bp - buf);            /* gelesene Bytes erh"ohen .. */
	  }

	cwrite (1, buf, bp - buf);                  /* Bytes schreiben ... */
	}
      return;
    }

  buf = (char *) xmalloc (nchars);
  do
    {
      /* Fill up the full buffer size from the input file.  */

      n_read = stdread (buf + n_buffered, nchars - n_buffered);
      if (n_read < 0)
	error (EXIT_FAILURE, errno, "%s", infile);

      n_buffered += n_read;
      if (n_buffered != nchars)
	eof = 1;

      /* Find where to end this chunk.  */
      bp = buf + n_buffered;
      if (n_buffered == nchars)
	{
	  while (bp > buf && bp[-1] != '\n')
	    bp--;
	}

      /* If chunk has no newlines, use all the chunk.  */
      if (bp == buf)
	bp = buf + n_buffered;

      /* Output the chars as one output file.  */
      cwrite (1, buf, bp - buf);

      /* Discard the chars we just output; move rest of chunk
	 down to be the start of the next chunk.  Source and
	 destination probably overlap.  */
      n_buffered -= bp - buf;
      if (n_buffered > 0)
	memmove (buf, bp, n_buffered);
    }
  while (!eof);
  free (buf);
}

int
main (int argc, char **argv)
{
  struct stat stat_buf;
  int num;			/* numeric argument from command line */
  enum
    {
      type_undef, type_bytes, type_byteslines, type_lines, type_digits
    } split_type = type_undef;
  int in_blk_size;		/* optimal block size of input file device */
  char *buf;			/* file i/o buffer */
  int accum = 0;
  char *outbase;
  int c;
  int digits_optind = 0;

  /* F"urs Mapping ben"otigte Variablen ... */

  off_t current_pos, end_pos;
  size_t infsize = 0;
  char *inbase = NULL;

  program_name = argv[0];
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);
 
  in_map = out_map = FALSE;

  /* Parse command line options.  */

  infile = "-";
  outbase = "x";

  while (1)
    {
      /* This is the argv-index of the option we will read next.  */
      int this_optind = optind ? optind : 1;
      long int tmp_long;

      c = getopt_long (argc, argv, "0123456789vb:l:C:", longopts, (int *) 0);
      if (c == EOF)
	break;

      switch (c)
	{
	case 0:
	  break;

	case 'b':
	  if (split_type != type_undef)
	    usage (2, _("cannot split in more than one way"));
	  split_type = type_bytes;
	  if (xstrtol (optarg, NULL, 10, &tmp_long, "bkm") != LONGINT_OK
	      || tmp_long < 0 || tmp_long > INT_MAX)
	    usage (2, _("invalid number of bytes"));
	  accum = (int) tmp_long;
	  break;

	case 'l':
	  if (split_type != type_undef)
	    usage (2, _("cannot split in more than one way"));
	  split_type = type_lines;
	  if (xstrtol (optarg, NULL, 10, &tmp_long, "") != LONGINT_OK
	      || tmp_long < 0 || tmp_long > INT_MAX)
	    usage (2, _("invalid number of lines"));
	  accum = (int) tmp_long;
	  break;

	case 'C':
	  if (split_type != type_undef)
	    usage (2, _("cannot split in more than one way"));
	  split_type = type_byteslines;
	  if (xstrtol (optarg, NULL, 10, &tmp_long, "bkm") != LONGINT_OK
	      || tmp_long < 0 ||  tmp_long > INT_MAX)
	    usage (2, _("invalid number of bytes"));
	  accum = (int) tmp_long;
	  break;

	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	  if (split_type != type_undef && split_type != type_digits)
	    usage (2, _("cannot split in more than one way"));
	  if (digits_optind != 0 && digits_optind != this_optind)
	    accum = 0;		/* More than one number given; ignore other. */
	  digits_optind = this_optind;
	  split_type = type_digits;
	  accum = accum * 10 + c - '0';
	  break;

	case 2:
	  verbose = 1;
	  break;

	default:
	  usage (2, (char *)0);
	}
    }

  if (show_version)
    {
      printf ("msplit - %s %s\n", GNU_PACKAGE, VERSION);
      exit (EXIT_SUCCESS);
    }

  if (show_help)
    usage (0, (char *)0);

  /* Handle default case.  */
  if (split_type == type_undef)
    {
      split_type = type_lines;
      accum = 1000;
    }

  if (accum < 1)
    usage (2, _("invalid number"));
  num = accum;

  /* Get out the filename arguments.  */

  if (optind < argc)
    infile = argv[optind++];

  if (optind < argc)
    outbase = argv[optind++];

  if (optind < argc)
    usage (2, _("too many arguments"));

  /* Open the input file.  */
  if (!strcmp (infile, "-"))
    input_desc = 0;
  else
    {
      input_desc = open (infile, O_RDONLY);
      if (input_desc < 0)
	error (EXIT_FAILURE, errno, "%s", infile);
    }

  /* No output file is open now.  */
  output_desc = -1;

  /* Copy the output file prefix so we can add suffixes to it.
     26**29 is certainly enough output files!  */

  outfile = xmalloc (strlen (outbase) + 30);
  strcpy (outfile, outbase);
  outfile_mid = outfile + strlen (outfile);
  outfile_end = outfile_mid + 2;
  memset (outfile_mid, 0, 30);
  outfile_mid[0] = 'a';
  outfile_mid[1] = 'a' - 1;  /* first call to next_file_name makes it an 'a' */

  /* Get the optimal block size of input device and make a buffer.  */
  /* Und Pr"ufung ob das input-file gemappt werden kann ... */

  if (fstat (input_desc, &stat_buf) < 0)
    error (EXIT_FAILURE, errno, "%s", infile);
  in_blk_size = ST_BLKSIZE (stat_buf);

  /* Dieser Test ist F"urs Mapping : Es wird getestet, ob die Datei 
   * grunds"atzlich map-bar ist; und zugleich wird die Gr"osse ermittelt. */
 
  if ( S_ISREG (stat_buf.st_mode)
      && (current_pos = lseek (input_desc, (off_t) 0, SEEK_CUR)) != -1
      && (end_pos = lseek (input_desc, (off_t) 0, SEEK_END)) != -1)
    { /* Mapping - Zweig. */
      off_t diff;
      infsize = (diff = end_pos - current_pos) < 0 ? 0 : diff;
      lseek (input_desc, (off_t) current_pos, SEEK_SET); /* R"ucksetzen !! */
      if ((infsize > 0 ) && 
          ((buf = mmap (NULL, infsize, PROT_READ, MAP_PRIVATE, \
                          input_desc, 0)) != NULL))
	{ 
	  in_map = TRUE;
	  fprintf(stderr,"%s : mapping !!! (fsize : %d)\n",
                  program_name,infsize);
	  in_blk_size = infsize;
	}
    }


  if ( ! in_map )  /* Normales Einlesen; Buffer wird reserviert .... */
      /* --- ori ---- */

    buf = xmalloc (in_blk_size + 1);

  switch (split_type)
    {
    case type_digits:
    case type_lines:
      lines_split (num, buf, in_blk_size);
      break;

    case type_bytes:
      bytes_split (num, buf, in_blk_size);
      break;

    case type_byteslines:
      /*      line_bytes_split (num);   */
      line_bytes_split (num, buf, in_blk_size);
      break;

    default:
      abort ();
    }

  if (close (input_desc) < 0)
    error (EXIT_FAILURE, errno, "%s", infile);
  if (output_desc >= 0 && close (output_desc) < 0)
    error (EXIT_FAILURE, errno, "%s", outfile);

  exit (EXIT_SUCCESS);
}
